import { Lock, Observable } from "@hibas123/utils";
import IDB from "@hibas123/lib-idb";
import { LoggingBase } from "@hibas123/logging";

export class HttpError extends Error {
   constructor(public status: number, public statusText: string) {
      super(statusText);
      console.log(statusText);
   }
}

export interface Position {
   date: Date;
   chapter: string;
   progress: number;
}

export interface VerySmallWork {
   _id: string;
   position: Position;
   creation_date: Date;
}

export interface BaseWork extends VerySmallWork {
   type: string;
   workid: string;

   title: string;
   summary: string;
   tags: string[];
   author: string;
   cover: string;
   new: boolean;
   folder: string;
   words: number;

   progress: number;
}

export interface BaseChapter {
   id: string;
   title: string;
   words: number;
}

export interface FullChapter extends BaseChapter {
   content: string;
}

export interface FullWork extends BaseWork {
   chapters: FullChapter[];
}

export interface SmallWork extends BaseWork {
   chapters: BaseChapter[];
}

interface ChapterData {
   _id: string;
   chapters: FullChapter[];
}

export enum WorkChangeTypes {
   create_edit,
   delete,
   position,
}

export interface IManuallyChapter {
   id: string;
   title: string;
   content: string;
}

export interface IManuallyEbook {
   workid: string;
   type: "manually";
   title: string;
   summary: string;
   tags: string[];
   author: string;
   cover: string;
   chapters: IManuallyChapter[];
}

const typeconv = ["manually", "epub", "wattpad", "ao3"];

// const config: { authserver: string, client_id: string } = (<any>window).config;
// if (!config) alert("Configuration not found, please reload!")

const ClientVersion = 1;

const upgrades = {
   upgradeType: (work: BaseWork) => {
      if (typeof work.type === "number") {
         work.type = typeconv[work.type] || "manually";
         return true;
      }
      return false;
   },
};

class EbookStorage {
   private idb = new IDB("eserver", ["works", "chapters"]);
   private chapterdb = this.idb.getStore<{ id: string; content: string }[]>(
      "chapters"
   );
   private workdb = this.idb.getStore<SmallWork>("works");

   private databaseLock = new Lock();

   private async upgrade() {
      const clientVersion = Number(localStorage.getItem("client_version") || 0);
      if (ClientVersion !== clientVersion) {
         const tx = await this.idb.transaction();
         switch (clientVersion) {
            case 0: {
               const all = await this.workdb.getAll(tx);
               await Promise.all(
                  all.map(async (elm: FullWork) => {
                     if (upgrades.upgradeType(elm)) {
                        await this.workdb.set(elm._id, elm);
                     }
                     return this.chapterdb.get(elm._id, tx).then((res) => {
                        if (!res) {
                           return this.chapterdb
                              .set(
                                 elm._id,
                                 elm.chapters.map((e) => ({
                                    id: e.id,
                                    content: e.content,
                                 })),
                                 tx
                              )
                              .then(() => {
                                 elm.chapters.forEach(
                                    (ch) => delete ch.content
                                 );
                                 return this.workdb.set(elm._id, elm, tx);
                              });
                        }
                     });
                  })
               );
            }
         }

         localStorage.setItem("client_version", String(ClientVersion));
      }
   }

   constructor() {
      const lock = this.databaseLock
         .getLock()
         .then(async (lock) => this.upgrade().then(() => lock.release()));
   }

   private getProgress(work: SmallWork) {
      // console.log("Get Progress", work);
      if (!work.position) return 0;
      let words =
         work.words ||
         work.chapters.reduce((prev, curr) => prev + curr.words, 0) ||
         0;
      let w = 0;
      work.chapters.find((e) => {
         let is = e.id === work.position.chapter;
         if (!is) {
            w += e.words;
         } else {
            w += work.position.progress * e.words;
         }
         return is;
      });
      return w / words;
   }

   async clean(): Promise<void> {
      await this.databaseLock.getLock().then((l) => l.release());
      return Promise.all([
         this.workdb.clear(),
         this.chapterdb.clear(),
      ]) as Promise<any>;
   }

   async getAll(): Promise<SmallWork[]> {
      await this.databaseLock.getLock().then((l) => l.release());
      return this.workdb.getAll();
   }

   async getWork(id: string): Promise<FullWork> {
      await this.databaseLock.getLock().then((l) => l.release());
      let [work, chapterData] = await Promise.all([
         this.workdb.get(id) as Promise<FullWork>,
         this.chapterdb.get(id),
      ]);

      work.chapters.forEach((chap) => {
         chap.content = chapterData.find((e) => e.id == chap.id).content;
      });

      return work;
   }

   async deleteWork(id: string): Promise<void> {
      await this.databaseLock.getLock().then((l) => l.release());
      const tx = await this.idb.transaction(this.workdb, this.chapterdb);
      await Promise.all([
         this.chapterdb.delete(id, tx),
         this.workdb.delete(id, tx),
      ]);
   }

   async saveWork(work: FullWork): Promise<SmallWork> {
      await this.databaseLock.getLock().then((l) => l.release());
      upgrades.upgradeType(work);
      let chapterData = work.chapters.map((chapter) => ({
         id: chapter.id,
         content: chapter.content,
      }));
      work.chapters.forEach((chapter) => delete chapter.content);
      const tx = await this.idb.transaction(this.workdb, this.chapterdb);
      await Promise.all([
         this.chapterdb.set(work._id, chapterData, tx),
         this.workdb.set(work._id, work, tx),
      ]);
      return work;
   }

   async savePosition(id: string, position: Position): Promise<SmallWork> {
      await this.databaseLock.getLock().then((l) => l.release());
      let work = await this.workdb.get(id);
      if (!work) throw new Error("Work not found");
      work.position = position;
      work.progress = this.getProgress(work);
      work.new = false;
      await this.workdb.set(id, work);
      return work;
   }
}

export default class EbookServerApi {
   public readonly logging = new LoggingBase({
      console: false,
      name: "ebookstore-client",
   });

   public readonly finishedLimit = 0.98;

   private syncLock = new Lock();
   private refreshToken: string | undefined = undefined;
   private jwt: string | undefined;
   // private _idb = new IDB("eserver", ["works", "chapters"]);
   // private chapterdb = this._idb.getStore<{ id: string, content: string }[]>("chapters");
   // private workdb = this._idb.getStore<SmallWork>("works");
   private storage = new EbookStorage();

   private workObersvableServer = new Observable<{
      type: WorkChangeTypes;
      work: SmallWork;
   }>(300);
   public workObservable = this.workObersvableServer.getPublicApi();

   private loginObeservableServer = new Observable<{
      need_login: boolean;
   }>();
   public loginObservable = this.loginObeservableServer.getPublicApi();

   private _needLogin: boolean = true;
   public get needLogin() {
      return this._needLogin;
   }

   public set needLogin(val) {
      if (val !== this._needLogin) {
         this._needLogin = val;
         this.loginObeservableServer.send({ need_login: val });
      }
   }
   public authLock = new Lock();

   get loginUrl() {
      return `${this.auth_server}/api/internal/oauth?client_id=${
         this.client_id
      }&redirect_uri=${encodeURIComponent(
         location.protocol + "//" + location.host
      )}`;
   }

   constructor(
      private auth_server: string,
      private client_id: string,
      private version: number
   ) {
      if (this.auth_server.endsWith("/"))
         this.auth_server = this.auth_server.slice(0, -1);

      const appVersion = Number(localStorage.getItem("app_version") || 0);

      if (appVersion !== version) {
         this.storage.clean().then(() => {
            localStorage.setItem("app_version", String(version));
         });
      }

      // if (version === 0) {
      //     this.legacy = true;
      // }

      this.needLogin = !localStorage.getItem("refreshtoken");
      this.init();
   }

   private async getJWT() {
      this.logging.log("Getting JWT");
      let res = await this.makeRequest(
         this.auth_server + "/api/oauth/jwt?refreshtoken=" + this.refreshToken,
         "GET",
         undefined,
         undefined,
         true
      );
      this.jwt = res.token;
   }

   private async init(nolock?: boolean) {
      this.logging.log("Running init");

      let lock = <any>undefined;
      if (!nolock) lock = await this.authLock.getLock();
      let token = localStorage.getItem("refreshtoken");
      let needLogin = false;
      if (token !== null && token !== "") {
         //ToDo just get JWT
         this.refreshToken = token;
         try {
            await this.getJWT();
            this.sync();
            setInterval(() => this.sync(), 60000);
         } catch (e) {
            if (e instanceof HttpError) {
               if (e.status === 400) {
                  //Bad Request
                  this.logging.log("Refresh token invalid, forward to login");
                  localStorage.removeItem("refreshtoken");
                  needLogin = true;
               }
            }
         }
      } else {
         needLogin = true;
      }
      this.needLogin = needLogin;

      if (!nolock) lock.release();
   }

   async getToken(code: string) {
      let lock = await this.authLock.getLock();
      let res = await this.makeRequest(
         "/api/get_refresh?code=" + code,
         "GET",
         undefined,
         undefined,
         true
      );
      localStorage.setItem("refreshtoken", res.token);
      await this.init(true);
      lock.release();
   }

   async logout() {
      await localStorage.clear();
      await this.storage.clean();
      this.jwt = undefined;
      this.refreshToken = undefined;
      this.needLogin = true;
   }

   async waitTillInitialized() {
      (await this.authLock.getLock()).release();
   }

   async makeRequest<T = any>(
      url: string,
      method: "POST" | "GET" | "PUT" | "DELETE" = "GET",
      body?: any,
      contenttype?: string,
      noauth?: boolean,
      failjwt?: boolean
   ): Promise<T> {
      this.logging.debug(
         "Making request: ",
         url,
         method,
         body,
         noauth,
         failjwt
      );
      let header: any = {
         "Content-Type": contenttype || "application/json",
      };

      if (contenttype === "noct") header = {};

      if (!noauth) {
         header.Authorization = this.jwt;
      }
      let bd;
      if (typeof body === "string") bd = body;
      else if (body instanceof Blob) bd = body;
      else bd = JSON.stringify(body);
      let res = await fetch(url, {
         headers: header,
         method: method,
         body: bd,
         credentials: "same-origin",
         mode: "cors",
      });

      const newJWT = async () => {
         await this.getJWT();
         return this.makeRequest(url, method, body, contenttype, noauth, true);
      };
      if (res.status === 418 && !failjwt) {
         //JWT is expired or invalid
         this.logging.log("JWT expired or invalid");
         return newJWT();
      }
      if (res.status !== 200) throw new HttpError(res.status, res.statusText);
      let data = await res.json();
      // console.log("Got data:", data)
      if (data.error) {
         if (data.status === 418 && !failjwt) {
            this.logging.log("JWT Expired");
            return newJWT();
         }
         throw new HttpError(data.status, data.error);
      }
      return data;
   }

   public getVersion() {
      return this.makeRequest("/version_hash", "GET").then((e) => e.version);
   }

   public async getJsonWebToken() {
      await this.getJWT();
      return this.jwt;
   }

   private fixCoverUrl(url: string) {
      if (!url) return undefined;
      if (!url.startsWith) return "/" + url;
      return url;
   }

   private async getVerySmallWorksRemote() {
      return (
         await this.makeRequest<{ works: VerySmallWork[] }>(
            "/api/works?small=true"
         )
      ).works;
   }

   private async getWorksRemote(): Promise<SmallWork[]> {
      return (await this.makeRequest<{ works: SmallWork[] }>("/api/works"))
         .works;
   }

   async getWorks(): Promise<SmallWork[]> {
      let works = (await this.storage.getAll()).map((e) => {
         if (e.cover) e.cover = this.fixCoverUrl(e.cover);
         return e;
      });
      return works;
   }

   private async getWorkRemote(id: string): Promise<FullWork> {
      let work = await await this.makeRequest("/api/works/" + id);
      upgrades.upgradeType(work);
      return work;
   }

   async getWork(id: string): Promise<FullWork> {
      let work = await this.storage.getWork(id);

      if (!work) work = await this.getWorkRemote(id);
      if (work) work.cover = this.fixCoverUrl(work.cover);

      return work;
   }

   async resolveUrl(
      url: string
   ): Promise<{ url: string; type: string; id: string }> {
      return this.makeRequest("/api/resolve?url=" + encodeURIComponent(url));
   }

   async addWorkFromUrl(input: string) {
      const { type, id } = await this.resolveUrl(input);

      if (type && id) return this.addWorkWithId(type, id);

      throw new Error("URL cannot be identified!");
   }

   async addEPub(file: Blob) {
      let type = this.version === 0 ? 1 : "epub";
      let res = await this.makeRequest<FullWork>(
         `/api/works?type=${type}`,
         "POST",
         file,
         "noct"
      );
      // await this.workdb.set(res._id, res);
      await this.storage.saveWork(res);
      this.workObersvableServer.send({
         type: WorkChangeTypes.create_edit,
         work: res,
      });
   }

   async addAO3(workid: string) {
      return this.addWorkWithId("ao3", workid);
   }

   async addWattpad(workid: string, pos?: string) {
      return this.addWorkWithId("wattpad", workid, pos);
   }

   async addWorkWithId(type: string, id: string, pos?: string): Promise<void> {
      let p = "";
      if (pos) p = `&position=${pos}`;
      let t;
      if (this.version === 0) {
         t = type === "wattpad" ? 2 : 3;
      } else {
         t = type;
      }
      let res = await this.makeRequest<FullWork>(
         `/api/works?type=${t}&workid=${id}`,
         "POST"
      );

      await this.storage.saveWork(res);
      this.workObersvableServer.send({
         type: WorkChangeTypes.create_edit,
         work: res,
      });
   }

   async addManualWork(work: IManuallyEbook) {
      let res = await this.makeRequest<FullWork>(
         `/api/works?type=manually`,
         "POST",
         JSON.stringify(work),
         "application/octet-stream"
      );

      this.storage.saveWork(res);
      this.workObersvableServer.send({
         type: WorkChangeTypes.create_edit,
         work: res,
      });
   }

   async deleteWork(id: string) {
      let res = await this.makeRequest<{ success: true }>(
         "/api/works/" + id,
         "DELETE"
      );
      if (res.success) {
         await this.storage.deleteWork(id);
      } else {
         throw new Error("Could not delete!");
      }
      return;
   }

   private async pushPositionRemote(id: string, position: Position) {
      return this.makeRequest(`/api/works/${id}/position`, "POST", position);
   }

   private async pushPositionLocal(
      id: string,
      position: Position
   ): Promise<SmallWork> {
      return this.storage.savePosition(id, position);
   }

   async pushPosition(id: string, position: Position) {
      let work = await this.pushPositionLocal(id, position);
      this.workObersvableServer.send({
         type: WorkChangeTypes.position,
         work: work,
      });
      return this.pushPositionRemote(id, position);
   }

   async sync() {
      //Another Sync process is running currently, so another one is not necessary
      if (this.syncLock.locked) return;
      this.logging.log("Starting sync");
      const lock = await this.syncLock.getLock();
      try {
         this.logging.debug("Fetching local");
         let locals = await this.storage.getAll();

         this.logging.debug("Fetching remote");
         let remotes = await this.getVerySmallWorksRemote();

         let towait: Promise<void>[] = [];
         let to_update = remotes.filter((rem) => {
            let loc = locals.find((e) => e._id === rem._id);
            if (!loc) return true;
            if (
               new Date(rem.creation_date).getTime() >
               new Date(loc.creation_date).getTime()
            )
               return true;
            else {
               //Check if position changed or need to be changed!
               if (!!loc.position || !!rem.position) {
                  if (
                     (!rem.position && loc.position) ||
                     (!loc.position && rem.position) ||
                     rem.position.chapter !== loc.position.chapter ||
                     rem.position.progress !== loc.position.progress
                  ) {
                     if (
                        !loc.position ||
                        (rem.position &&
                           new Date(rem.position.date).getTime() >=
                              new Date(loc.position.date).getTime())
                     ) {
                        // remote is newer
                        this.logging.debug("Updating local position:", rem._id);
                        // loc.position = rem.position;
                        // loc.progress = this.getProgress(loc);
                        // loc.new = false;
                        towait.push(
                           this.storage
                              .savePosition(rem._id, rem.position)
                              .then(() =>
                                 this.workObersvableServer.send({
                                    type: WorkChangeTypes.position,
                                    work: loc,
                                 })
                              )
                        );
                     } else {
                        this.logging.debug(
                           "Updating remote position:",
                           rem._id
                        );
                        towait.push(
                           this.pushPositionRemote(loc._id, loc.position)
                        );
                     }
                  }
               }
            }
            return false;
         });

         await Promise.all(towait);

         const download = (async () => {
            let work: VerySmallWork = to_update.pop();
            while (work) {
               try {
                  this.logging.debug("Downloading and storing work", work);
                  let w = await this.getWorkRemote(work._id);

                  await this.storage.saveWork(w);
                  this.workObersvableServer.send({
                     type: WorkChangeTypes.create_edit,
                     work: w,
                  });
               } catch (e) {
                  this.logging.error(e);
               }
               work = to_update.pop();
            }
         }).bind(this);

         if (to_update.length > 0) this.logging.debug("Updating local works");
         towait = [];
         for (let i = 0; i < 3; i++) {
            towait.push(download());
         }
         Promise.all(towait);

         let to_remove = locals.filter(
            (a) => !remotes.find((b) => b._id === a._id)
         );
         for (let work of to_remove) {
            this.logging.debug("Deleting work", work);
            await this.storage.deleteWork(work._id);
            this.workObersvableServer.send({
               type: WorkChangeTypes.delete,
               work: work,
            });
         }
      } catch (e) {
         this.logging.error(e);
      } finally {
         lock.release();
         this.logging.log("Sync finished");
      }
   }
}
